@@ -53,7 +53,7 @@
 #include "linux-dmabuf-unstable-v1-client-protocol.h"
 
 struct xwl_gbm_private {
-    dev_t device;
+    drmDevice *device;
     char *device_name;
     struct gbm_device *gbm;
     struct wl_drm *drm;
@@ -69,6 +69,7 @@ struct xwl_pixmap {
     EGLImage image;
     unsigned int texture;
     struct gbm_bo *bo;
+    Bool implicit_modifier;
 };
 
 static DevPrivateKeyRec xwl_gbm_private_key;
@@ -116,7 +117,8 @@ is_device_path_render_node (const char *device_path)
 
 static PixmapPtr
 xwl_glamor_gbm_create_pixmap_for_bo(ScreenPtr screen, struct gbm_bo *bo,
-                                    int depth)
+                                    int depth,
+                                    Bool implicit_modifier)
 {
     PixmapPtr pixmap;
     struct xwl_pixmap *xwl_pixmap;
@@ -188,6 +190,7 @@ xwl_glamor_gbm_create_pixmap_for_bo(ScreenPtr screen, struct gbm_bo *bo,
     xwl_glamor_egl_make_current(xwl_screen);
     xwl_pixmap->bo = bo;
     xwl_pixmap->buffer = NULL;
+    xwl_pixmap->implicit_modifier = implicit_modifier;
 
 #ifdef GBM_BO_FD_FOR_PLANE
     if (xwl_gbm->dmabuf_capable) {
@@ -286,6 +289,7 @@ xwl_glamor_gbm_create_pixmap(ScreenPtr screen,
          hint == CREATE_PIXMAP_USAGE_SHARED ||
          (xwl_screen->rootless && hint == 0))) {
         uint32_t format = gbm_format_for_depth(depth);
+        Bool implicit = FALSE;
 
 #ifdef GBM_BO_WITH_MODIFIERS
         if (xwl_gbm->dmabuf_capable) {
@@ -293,18 +297,21 @@ xwl_glamor_gbm_create_pixmap(ScreenPtr screen,
             uint64_t *modifiers = NULL;
 
             xwl_glamor_get_modifiers(screen, format, &num_modifiers, &modifiers);
-            bo = gbm_bo_create_with_modifiers(xwl_gbm->gbm, width, height,
-                                              format, modifiers, num_modifiers);
+
+            if (num_modifiers > 0)
+                bo = gbm_bo_create_with_modifiers(xwl_gbm->gbm, width, height,
+                                                  format, modifiers, num_modifiers);
             free(modifiers);
         }
 #endif
         if (bo == NULL) {
+            implicit = TRUE;
             bo = gbm_bo_create(xwl_gbm->gbm, width, height, format,
                                GBM_BO_USE_RENDERING);
         }
 
         if (bo) {
-            pixmap = xwl_glamor_gbm_create_pixmap_for_bo(screen, bo, depth);
+            pixmap = xwl_glamor_gbm_create_pixmap_for_bo(screen, bo, depth, implicit);
 
             if (!pixmap) {
                 gbm_bo_destroy(bo);
@@ -345,42 +352,24 @@ static const struct wl_buffer_listener xwl_glamor_gbm_buffer_listener = {
     xwl_pixmap_buffer_release_cb,
 };
 
-static struct wl_buffer *
-xwl_glamor_gbm_get_wl_buffer_for_pixmap(PixmapPtr pixmap)
+#ifdef GBM_BO_WITH_MODIFIERS
+static Bool
+init_buffer_params_with_modifiers(struct xwl_pixmap *xwl_pixmap,
+                                  uint64_t          *modifier,
+                                  int               *num_planes,
+                                  int               *prime_fds,
+                                  uint32_t          *strides,
+                                  uint32_t          *offsets)
 {
-    struct xwl_screen *xwl_screen = xwl_screen_get(pixmap->drawable.pScreen);
-    struct xwl_pixmap *xwl_pixmap = xwl_pixmap_get(pixmap);
-    struct xwl_gbm_private *xwl_gbm = xwl_gbm_get(xwl_screen);
-    unsigned short width = pixmap->drawable.width;
-    unsigned short height = pixmap->drawable.height;
-    uint32_t format;
-    int num_planes;
-    int prime_fds[4];
-    uint32_t strides[4];
-    uint32_t offsets[4];
-    uint64_t modifier;
-    int i;
 #ifndef GBM_BO_FD_FOR_PLANE
     int32_t first_handle;
 #endif
+    int i;
 
-    if (xwl_pixmap == NULL)
-       return NULL;
+    *num_planes = gbm_bo_get_plane_count(xwl_pixmap->bo);
+    *modifier = gbm_bo_get_modifier(xwl_pixmap->bo);
 
-    if (xwl_pixmap->buffer) {
-        /* Buffer already exists. */
-        return xwl_pixmap->buffer;
-    }
-
-    if (!xwl_pixmap->bo)
-       return NULL;
-
-    format = wl_drm_format_for_depth(pixmap->drawable.depth);
-
-#ifdef GBM_BO_WITH_MODIFIERS
-    num_planes = gbm_bo_get_plane_count(xwl_pixmap->bo);
-    modifier = gbm_bo_get_modifier(xwl_pixmap->bo);
-    for (i = 0; i < num_planes; i++) {
+    for (i = 0; i < *num_planes; i++) {
 #ifdef GBM_BO_FD_FOR_PLANE
         prime_fds[i] = gbm_bo_get_fd_for_plane(xwl_pixmap->bo, i);
 #else
@@ -403,20 +392,85 @@ xwl_glamor_gbm_get_wl_buffer_for_pixmap(PixmapPtr pixmap)
         if (prime_fds[i] == -1) {
             while (--i >= 0)
                 close(prime_fds[i]);
-            return NULL;
+            return FALSE;
         }
         strides[i] = gbm_bo_get_stride_for_plane(xwl_pixmap->bo, i);
         offsets[i] = gbm_bo_get_offset(xwl_pixmap->bo, i);
     }
-#else
-    num_planes = 1;
-    modifier = DRM_FORMAT_MOD_INVALID;
+
+    return TRUE;
+}
+#endif
+
+static Bool
+init_buffer_params_fallback(struct xwl_pixmap *xwl_pixmap,
+                            uint64_t          *modifier,
+                            int               *num_planes,
+                            int               *prime_fds,
+                            uint32_t          *strides,
+                            uint32_t          *offsets)
+{
+    *num_planes = 1;
+    *modifier = DRM_FORMAT_MOD_INVALID;
     prime_fds[0] = gbm_bo_get_fd(xwl_pixmap->bo);
     if (prime_fds[0] == -1)
-        return NULL;
+        return FALSE;
+
     strides[0] = gbm_bo_get_stride(xwl_pixmap->bo);
     offsets[0] = 0;
+
+    return TRUE;
+}
+
+static struct wl_buffer *
+xwl_glamor_gbm_get_wl_buffer_for_pixmap(PixmapPtr pixmap)
+{
+    struct xwl_screen *xwl_screen = xwl_screen_get(pixmap->drawable.pScreen);
+    struct xwl_pixmap *xwl_pixmap = xwl_pixmap_get(pixmap);
+    struct xwl_gbm_private *xwl_gbm = xwl_gbm_get(xwl_screen);
+    unsigned short width = pixmap->drawable.width;
+    unsigned short height = pixmap->drawable.height;
+    uint32_t format;
+    int num_planes;
+    int prime_fds[4];
+    uint32_t strides[4];
+    uint32_t offsets[4];
+    uint64_t modifier;
+    int i;
+
+    if (xwl_pixmap == NULL)
+       return NULL;
+
+    if (xwl_pixmap->buffer) {
+        /* Buffer already exists. */
+        return xwl_pixmap->buffer;
+    }
+
+    if (!xwl_pixmap->bo)
+       return NULL;
+
+    format = wl_drm_format_for_depth(pixmap->drawable.depth);
+
+#ifdef GBM_BO_WITH_MODIFIERS
+    if (!xwl_pixmap->implicit_modifier) {
+        if (!init_buffer_params_with_modifiers(xwl_pixmap,
+                                               &modifier,
+                                               &num_planes,
+                                               prime_fds,
+                                               strides,
+                                               offsets))
+            return NULL;
+    } else
 #endif
+    {
+        if (!init_buffer_params_fallback(xwl_pixmap,
+                                         &modifier,
+                                         &num_planes,
+                                         prime_fds,
+                                         strides,
+                                         offsets))
+            return NULL;
+    }
 
     if (xwl_screen->dmabuf &&
         xwl_glamor_is_modifier_supported(xwl_screen, format, modifier)) {
@@ -433,7 +487,7 @@ xwl_glamor_gbm_get_wl_buffer_for_pixmap(PixmapPtr pixmap)
            zwp_linux_buffer_params_v1_create_immed(params, width, height,
                                                    format, 0);
         zwp_linux_buffer_params_v1_destroy(params);
-    } else if (num_planes == 1) {
+    } else if (num_planes == 1 && modifier == DRM_FORMAT_MOD_INVALID) {
         xwl_pixmap->buffer =
             wl_drm_create_prime_buffer(xwl_gbm->drm, prime_fds[0], width, height,
                                        format,
@@ -446,8 +500,9 @@ xwl_glamor_gbm_get_wl_buffer_for_pixmap(PixmapPtr pixmap)
         close(prime_fds[i]);
 
     /* Add our listener now */
-    wl_buffer_add_listener(xwl_pixmap->buffer,
-                           &xwl_glamor_gbm_buffer_listener, pixmap);
+    if (xwl_pixmap->buffer)
+        wl_buffer_add_listener(xwl_pixmap->buffer,
+                               &xwl_glamor_gbm_buffer_listener, pixmap);
 
     return xwl_pixmap->buffer;
 }
@@ -459,6 +514,7 @@ xwl_glamor_gbm_cleanup(struct xwl_screen *xwl_screen)
 
     if (xwl_gbm->device_name)
         free(xwl_gbm->device_name);
+    drmFreeDevice(&xwl_gbm->device);
     if (xwl_gbm->drm_fd)
         close(xwl_gbm->drm_fd);
     if (xwl_gbm->drm)
@@ -578,6 +634,7 @@ glamor_pixmap_from_fds(ScreenPtr screen, CARD8 num_fds, const int *fds,
     struct gbm_bo *bo = NULL;
     PixmapPtr pixmap;
     int i;
+    Bool implicit = FALSE;
 
     if (width == 0 || height == 0 || num_fds == 0 ||
         depth < 15 || bpp != BitsPerPixel(depth) ||
@@ -611,6 +668,7 @@ glamor_pixmap_from_fds(ScreenPtr screen, CARD8 num_fds, const int *fds,
        data.format = gbm_format_for_depth(depth);
        bo = gbm_bo_import(xwl_gbm->gbm, GBM_BO_IMPORT_FD, &data,
                           GBM_BO_USE_RENDERING);
+       implicit = TRUE;
     } else {
        goto error;
     }
@@ -618,7 +676,7 @@ glamor_pixmap_from_fds(ScreenPtr screen, CARD8 num_fds, const int *fds,
     if (bo == NULL)
        goto error;
 
-    pixmap = xwl_glamor_gbm_create_pixmap_for_bo(screen, bo, depth);
+    pixmap = xwl_glamor_gbm_create_pixmap_for_bo(screen, bo, depth, implicit);
     if (pixmap == NULL) {
        gbm_bo_destroy(bo);
        goto error;
@@ -783,7 +841,6 @@ xwl_drm_handle_device(void *data, struct wl_drm *drm, const char *device)
    struct xwl_gbm_private *xwl_gbm = xwl_gbm_get(xwl_screen);
    drm_magic_t magic;
    char *render_node_path = NULL;
-   struct stat stat;
 
    if (!is_device_path_render_node(device))
        render_node_path = get_render_node_path(device);
@@ -806,12 +863,11 @@ xwl_drm_handle_device(void *data, struct wl_drm *drm, const char *device)
        return;
    }
 
-   if (fstat(xwl_gbm->drm_fd, &stat)) {
-       ErrorF("wayland-egl: Could not stat file %s (%s)\n",
-              xwl_gbm->device_name, strerror(errno));
+   if (drmGetDevice2(xwl_gbm->drm_fd, 0, &xwl_gbm->device) != 0) {
+       ErrorF("wayland-egl: Could not fetch DRM device %s\n",
+              xwl_gbm->device_name);
        return;
    }
-   xwl_gbm->device = stat.st_rdev;
 
    if (drmGetNodeTypeFromFd(xwl_gbm->drm_fd) == DRM_NODE_RENDER) {
        xwl_gbm->fd_render_node = 1;
@@ -1110,7 +1166,7 @@ error:
     return FALSE;
 }
 
-static dev_t xwl_gbm_get_main_device(struct xwl_screen *xwl_screen)
+static drmDevice *xwl_gbm_get_main_device(struct xwl_screen *xwl_screen)
 {
     struct xwl_gbm_private *xwl_gbm = xwl_gbm_get(xwl_screen);
 
